import pytest
from pytest_mock import MockerFixture

from mrok.ziti.api import TagsType
from mrok.ziti.services import (
    ConfigTypeNotFoundError,
    ProxyIdentityNotFoundError,
    ServiceAlreadyRegisteredError,
    ServiceNotFoundError,
    register_extension,
    unregister_extension,
)
from tests.conftest import SettingsFactory


@pytest.mark.asyncio
async def test_register_extension(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = {"id": "proxy_identity_id"}
    mocked_api.search_config_type.return_value = {"id": "config_type_id"}
    mocked_api.search_config.return_value = None
    mocked_api.create_config.return_value = "config_id"
    mocked_api.search_service.return_value = None
    mocked_api.create_service.return_value = "service_id"
    mocked_api.get_service.return_value = {"id": "service_id"}
    mocked_api.search_service_policy.return_value = None
    mocked_api.search_service_router_policy.return_value = None

    tags: TagsType = {"tag": "my-tag"}

    await register_extension(settings, mocked_api, "EXT-1234", tags)

    mocked_api.search_identity.assert_awaited_once_with(settings.proxy.identity)
    mocked_api.search_config_type.assert_awaited_once_with(f"{settings.proxy.mode}.proxy.v1")
    mocked_api.search_config.assert_awaited_once_with("ext-1234")
    mocked_api.create_config.assert_awaited_once_with(
        "ext-1234",
        "config_type_id",
        tags=tags,
    )
    mocked_api.search_service.assert_called_once_with("ext-1234")
    mocked_api.create_service.assert_called_once_with("ext-1234", "config_id", tags=tags)
    mocked_api.get_service.assert_awaited_once_with("service_id")
    mocked_api.search_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial"
    )
    mocked_api.create_dial_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial",
        "service_id",
        "proxy_identity_id",
        tags=tags,
    )
    mocked_api.search_service_router_policy.assert_awaited_once_with("ext-1234")
    mocked_api.create_service_router_policy.assert_awaited_once_with(
        "ext-1234",
        "service_id",
        tags=tags,
    )


@pytest.mark.asyncio
async def test_register_extension_no_proxy_identity(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = None

    with pytest.raises(ProxyIdentityNotFoundError) as cv:
        await register_extension(settings, mocked_api, "EXT-1234", None)
    assert str(cv.value) == f"Identity for proxy `{settings.proxy.identity}` not found."


@pytest.mark.asyncio
async def test_register_extension_no_config_type(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = {"id": "proxy_identity_id"}
    mocked_api.search_config_type.return_value = None

    with pytest.raises(ConfigTypeNotFoundError) as cv:
        await register_extension(settings, mocked_api, "EXT-1234", None)
    assert str(cv.value) == f"Config type `{settings.proxy.mode}.proxy.v1` not found."


@pytest.mark.asyncio
async def test_register_extension_config_exists(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = {"id": "proxy_identity_id"}
    mocked_api.search_config_type.return_value = {"id": "config_type_id"}
    mocked_api.search_config.return_value = {"id": "config_id"}
    mocked_api.search_service.return_value = None
    mocked_api.create_service.return_value = "service_id"
    mocked_api.get_service.return_value = {"id": "service_id"}
    mocked_api.search_service_policy.return_value = None
    mocked_api.search_service_router_policy.return_value = None

    tags: TagsType = {"tag": "my-tag"}

    await register_extension(settings, mocked_api, "EXT-1234", tags)

    mocked_api.search_identity.assert_awaited_once_with(settings.proxy.identity)
    mocked_api.search_config_type.assert_awaited_once_with(f"{settings.proxy.mode}.proxy.v1")
    mocked_api.search_config.assert_awaited_once_with("ext-1234")
    mocked_api.create_config.assert_not_awaited()
    mocked_api.search_service.assert_called_once_with("ext-1234")
    mocked_api.create_service.assert_called_once_with("ext-1234", "config_id", tags=tags)
    mocked_api.search_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial"
    )
    mocked_api.create_dial_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial",
        "service_id",
        "proxy_identity_id",
        tags=tags,
    )
    mocked_api.search_service_router_policy.assert_awaited_once_with("ext-1234")
    mocked_api.create_service_router_policy.assert_awaited_once_with(
        "ext-1234",
        "service_id",
        tags=tags,
    )


@pytest.mark.asyncio
async def test_register_extension_service_exists(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = {"id": "proxy_identity_id"}
    mocked_api.search_config_type.return_value = {"id": "config_type_id"}
    mocked_api.search_config.return_value = {"id": "config_id"}
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_policy.return_value = None
    mocked_api.search_service_router_policy.return_value = None

    tags: TagsType = {"tag": "my-tag"}

    await register_extension(settings, mocked_api, "EXT-1234", tags)

    mocked_api.search_identity.assert_awaited_once_with(settings.proxy.identity)
    mocked_api.search_config_type.assert_awaited_once_with(f"{settings.proxy.mode}.proxy.v1")
    mocked_api.search_config.assert_awaited_once_with("ext-1234")
    mocked_api.create_config.assert_not_awaited()
    mocked_api.search_service.assert_called_once_with("ext-1234")
    mocked_api.create_service.assert_not_awaited()
    mocked_api.get_service.assert_not_awaited()
    mocked_api.search_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial"
    )
    mocked_api.create_dial_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial",
        "service_id",
        "proxy_identity_id",
        tags=tags,
    )
    mocked_api.search_service_router_policy.assert_awaited_once_with("ext-1234")
    mocked_api.create_service_router_policy.assert_awaited_once_with(
        "ext-1234",
        "service_id",
        tags=tags,
    )


@pytest.mark.asyncio
async def test_register_extension_dial_policy_exists(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = {"id": "proxy_identity_id"}
    mocked_api.search_config_type.return_value = {"id": "config_type_id"}
    mocked_api.search_config.return_value = {"id": "config_id"}
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_policy.return_value = {"id": "policy_id"}
    mocked_api.search_service_router_policy.return_value = None

    tags: TagsType = {"tag": "my-tag"}

    await register_extension(settings, mocked_api, "EXT-1234", tags)

    mocked_api.search_identity.assert_awaited_once_with(settings.proxy.identity)
    mocked_api.search_config_type.assert_awaited_once_with(f"{settings.proxy.mode}.proxy.v1")
    mocked_api.search_config.assert_awaited_once_with("ext-1234")
    mocked_api.create_config.assert_not_awaited()
    mocked_api.search_service.assert_called_once_with("ext-1234")
    mocked_api.create_service.assert_not_awaited()
    mocked_api.get_service.assert_not_awaited()
    mocked_api.search_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial"
    )
    mocked_api.create_dial_service_policy.assert_not_awaited()
    mocked_api.search_service_router_policy.assert_awaited_once_with("ext-1234")
    mocked_api.create_service_router_policy.assert_awaited_once_with(
        "ext-1234",
        "service_id",
        tags=tags,
    )


@pytest.mark.asyncio
async def test_register_extension_router_policy_exists(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_identity.return_value = {"id": "proxy_identity_id"}
    mocked_api.search_config_type.return_value = {"id": "config_type_id"}
    mocked_api.search_config.return_value = {"id": "config_id"}
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_policy.return_value = {"id": "policy_id"}
    mocked_api.search_service_router_policy.return_value = {"id": "policy_id"}

    tags: TagsType = {"tag": "my-tag"}

    with pytest.raises(ServiceAlreadyRegisteredError) as cv:
        await register_extension(settings, mocked_api, "EXT-1234", tags)
    assert str(cv.value) == "Extension `EXT-1234` already registered."

    mocked_api.search_identity.assert_awaited_once_with(settings.proxy.identity)
    mocked_api.search_config_type.assert_awaited_once_with(f"{settings.proxy.mode}.proxy.v1")
    mocked_api.search_config.assert_awaited_once_with("ext-1234")
    mocked_api.create_config.assert_not_awaited()
    mocked_api.search_service.assert_called_once_with("ext-1234")
    mocked_api.create_service.assert_not_awaited()
    mocked_api.get_service.assert_not_awaited()
    mocked_api.search_service_policy.assert_called_once_with(
        f"ext-1234:{settings.proxy.identity}:dial"
    )
    mocked_api.create_dial_service_policy.assert_not_awaited()
    mocked_api.search_service_router_policy.assert_awaited_once_with("ext-1234")
    mocked_api.create_service_router_policy.assert_not_awaited()


@pytest.mark.asyncio
async def test_unregister_extension(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_router_policy.return_value = {"id": "router_policy_id"}
    mocked_api.search_service_policy.return_value = {"id": "service_policy_id"}
    mocked_api.search_config.return_value = {"id": "config_id"}

    await unregister_extension(settings, mocked_api, "EXT-1234")

    mocked_api.delete_service_router_policy.assert_awaited_once_with("router_policy_id")
    mocked_api.delete_service_policy.assert_awaited_once_with("service_policy_id")
    mocked_api.delete_config.assert_awaited_once_with("config_id")
    mocked_api.delete_service.assert_awaited_once_with("service_id")


@pytest.mark.asyncio
async def test_unregister_extension_not_found(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_service.return_value = None

    with pytest.raises(ServiceNotFoundError) as cv:
        await unregister_extension(settings, mocked_api, "EXT-1234")
    assert str(cv.value) == "Extension `EXT-1234` not found."


@pytest.mark.asyncio
async def test_unregister_extension_router_policy_doesnt_exist(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_router_policy.return_value = None
    mocked_api.search_service_policy.return_value = {"id": "service_policy_id"}
    mocked_api.search_config.return_value = {"id": "config_id"}

    await unregister_extension(settings, mocked_api, "EXT-1234")

    mocked_api.delete_service_router_policy.assert_not_awaited()
    mocked_api.delete_service_policy.assert_awaited_once_with("service_policy_id")
    mocked_api.delete_config.assert_awaited_once_with("config_id")
    mocked_api.delete_service.assert_awaited_once_with("service_id")


@pytest.mark.asyncio
async def test_unregister_extension_service_policy_doesnt_exist(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_router_policy.return_value = None
    mocked_api.search_service_policy.return_value = None
    mocked_api.search_config.return_value = {"id": "config_id"}

    await unregister_extension(settings, mocked_api, "EXT-1234")

    mocked_api.delete_service_router_policy.assert_not_awaited()
    mocked_api.delete_service_policy.assert_not_awaited()
    mocked_api.delete_config.assert_awaited_once_with("config_id")
    mocked_api.delete_service.assert_awaited_once_with("service_id")


@pytest.mark.asyncio
async def test_unregister_extension_config_doesnt_exist(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.search_service.return_value = {"id": "service_id"}
    mocked_api.search_service_router_policy.return_value = None
    mocked_api.search_service_policy.return_value = None
    mocked_api.search_config.return_value = None

    await unregister_extension(settings, mocked_api, "EXT-1234")

    mocked_api.delete_service_router_policy.assert_not_awaited()
    mocked_api.delete_service_policy.assert_not_awaited()
    mocked_api.delete_config.assert_not_awaited()
    mocked_api.delete_service.assert_awaited_once_with("service_id")



import typer

from mrok.cli.commands.admin.register import extensions, instances

app = typer.Typer(name="register", help="Register resources into OpenZiti.")

extensions.register(app)
instances.register(app)



import pytest
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.x509.oid import NameOID
from pytest_mock import MockerFixture

from mrok.ziti import pki


@pytest.mark.asyncio
async def test_get_ca_certificates(mocker: MockerFixture):
    mocker.patch.object(pki, "_ca_certificates", None)
    mgmt_api = mocker.AsyncMock()
    mgmt_api.fetch_ca_certificates.return_value = (
        "MIIDNAYJKoZIhvcNAQcCoIIDJTCCAyECAQExADALBgkqhkiG9w0BBwGgggMJMIIDBTCCAe2gAwIB\n"
        "AgIUIPs0bOvdiihzrzOM/W5UW2gw+MkwDQYJKoZIhvcNAQELBQAwEjEQMA4GA1UEAwwHVGVzdCBD\n"
        "QTAeFw0yNTA5MjYwOTE0MDBaFw0yNTA5MjcwOTE0MDBaMBIxEDAOBgNVBAMMB1Rlc3QgQ0EwggEi\n"
        "MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDTOPLeCYvA1SqT7t9GqhD0K3/lNl4s0zpdYUxT\n"
        "POHPsR08Vol+87hGLvbKmUAI7+hSU9oBSIDwnuBnpvOoGTH9VYPnN/hxEVEbMmFkKWP+FwZxQn0a\n"
        "Bfb+iqENvwIij8tvq8nmrs3hCWGG1dgIr1ZnR5nWr+GWveCJpHcH0xXaOG2mZscw0VZMaFbc4zan\n"
        "5Us1hHBkYDSDSVwTT4C2GsaxIiwn26IXpjj1bRa1ffr7SdauJ78HNV/DVpCMOG6Kz1K0x6luv0HY\n"
        "OtXQvYkkhy7R5Ir9Li5LIgdaeLYnqKMkZVLO/JcMKTcZAiA7dNL5xSCEIKQmY9asy8n2bBHVks4R\n"
        "AgMBAAGjUzBRMB0GA1UdDgQWBBSznFqdu3fW0gIj9DkR/WYaOYDjFDAfBgNVHSMEGDAWgBSznFqd\n"
        "u3fW0gIj9DkR/WYaOYDjFDAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3DQEBCwUAA4IBAQBATWXe\n"
        "uz+4msC56Vc+rRsR1G7sAATexIm7vy/4Ou2vUmRb5UpsZvrgCRYRC7B1a0o2CwtaC1skgVResa1f\n"
        "Zhrk/b2Pn+Sc27pXxbjE1lA9T0IDUmRvEmPX/uBClOrF29EVeR9BnnBBAi1i2QduL/yLz8rgWMrb\n"
        "pthvNcpms0W9FlggEWm0ee7vhTLpft9eK4iA391IqZVt1N5CWMjk6pwvTc/6tqDUtKtpjgGh0NU4\n"
        "U6ET6dk3bqM4CNuLcj5tyg+bMTuxmViFC0qXRLtAQHSaYEF33aDraf6Uvv0KUywVsxn+4BTPpeuN\n"
        "gUjmhg6bpPvcC8waUDXINblhaDy6Z+gBMQA="
    )
    certs = await pki.get_ca_certificates(mgmt_api)

    assert certs.startswith("-----BEGIN CERTIFICATE-----")


@pytest.mark.asyncio
async def test_get_ca_certificates_already_fetched(mocker: MockerFixture):
    mocker.patch.object(pki, "_ca_certificates", "-----BEGIN CERTIFICATE-----")
    mgmt_api = mocker.AsyncMock()
    certs = await pki.get_ca_certificates(mgmt_api)
    mgmt_api.fetch_ca_certificates.assert_not_awaited()
    assert certs.startswith("-----BEGIN CERTIFICATE-----")


def test_generate_key_and_csr_subject_and_key_size():
    identity = "test-id"
    key_pem, csr_pem = pki.generate_key_and_csr(identity)

    # parse key
    private_key = serialization.load_pem_private_key(key_pem.encode(), password=None)
    assert private_key.key_size == 4096

    # parse csr
    csr = x509.load_pem_x509_csr(csr_pem.encode())
    subject = csr.subject

    def get(oid):
        return subject.get_attributes_for_oid(oid)[0].value

    assert get(NameOID.COUNTRY_NAME) == "CH"
    assert get(NameOID.ORGANIZATION_NAME) == "SoftwareOne"
    assert get(NameOID.ORGANIZATIONAL_UNIT_NAME) == "Marketplace Platform"
    assert get(NameOID.COMMON_NAME) == identity



import logging
from typing import Annotated

import httpx
import jwt
from fastapi import Depends, HTTPException, Request, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer

from mrok.controller.dependencies.conf import AppSettings

logger = logging.getLogger("mrok.controller")

UNAUTHORIZED_EXCEPTION = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized."
)


class JWTCredentials(HTTPAuthorizationCredentials):
    pass


class JWTBearer(HTTPBearer):
    def __init__(self):
        super().__init__(auto_error=False)

    async def __call__(self, request: Request) -> JWTCredentials:
        credentials = await super().__call__(request)
        if not credentials:
            raise UNAUTHORIZED_EXCEPTION
        try:
            return JWTCredentials(
                scheme=credentials.scheme,
                credentials=credentials.credentials,
            )
        except jwt.InvalidTokenError:
            raise UNAUTHORIZED_EXCEPTION


async def authenticate(
    settings: AppSettings,
    credentials: Annotated[JWTCredentials, Depends(JWTBearer())],
):
    async with httpx.AsyncClient(
        timeout=httpx.Timeout(
            connect=0.25,
            read=settings.auth.read_timeout,
            write=2.0,
            pool=5.0,
        ),
    ) as client:
        try:
            config_resp = await client.get(settings.auth.openid_config_url)
            config_resp.raise_for_status()
            config = config_resp.json()
            issuer = config["issuer"]
            jwks_uri = config["jwks_uri"]

            jwks_resp = await client.get(jwks_uri)
            jwks_resp.raise_for_status()
            jwks = jwks_resp.json()

            header = jwt.get_unverified_header(credentials.credentials)
            kid = header["kid"]

            key_data = next((k for k in jwks["keys"] if k["kid"] == kid), None)
        except Exception:
            logger.exception("Error fetching openid-config/jwks")
            raise UNAUTHORIZED_EXCEPTION
    if key_data is None:
        logger.error("Key ID not found in JWKS")
        raise UNAUTHORIZED_EXCEPTION

    try:
        payload = jwt.decode(
            credentials.credentials,
            jwt.PyJWK(key_data),
            algorithms=[header["alg"]],
            issuer=issuer,
            audience=settings.auth.audience,
        )
        return payload
    except jwt.InvalidKeyError as e:
        logger.error(f"Invalid jwt token: {e}")
        raise UNAUTHORIZED_EXCEPTION
    except jwt.InvalidTokenError as e:
        logger.error(f"Invalid jwt token: {e}")
        raise UNAUTHORIZED_EXCEPTION



import jwt
import pytest
from asgi_lifespan import LifespanManager
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from tests.conftest import SettingsFactory


@pytest.mark.asyncio
async def test_no_token(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
):
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
    ) as client:
        response = await client.get("/extensions")
        assert response.status_code == 401


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_header", ["Bearer invalid", "invalid", ""])
async def test_invalid_token(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    auth_header: str,
):
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
        headers={"Authorization": auth_header},
    ) as client:
        response = await client.get("/extensions")
        assert response.status_code == 401


@pytest.mark.asyncio
async def test_invalid_openid_config_url(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    httpx_mock: HTTPXMock,
    settings_factory: SettingsFactory,
    jwt_token: str,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.auth.openid_config_url,
        status_code=404,
    )
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
    ) as client:
        response = await client.get("/extensions", headers={"Authorization": f"Bearer {jwt_token}"})
        assert response.status_code == 401


@pytest.mark.asyncio
async def test_invalid_openid_config_data(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    httpx_mock: HTTPXMock,
    settings_factory: SettingsFactory,
    jwt_token: str,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.auth.openid_config_url,
        json={},
    )
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
    ) as client:
        response = await client.get("/extensions", headers={"Authorization": f"Bearer {jwt_token}"})
        assert response.status_code == 401


@pytest.mark.asyncio
async def test_invalid_jwks_url(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    httpx_mock: HTTPXMock,
    settings_factory: SettingsFactory,
    jwt_token: str,
    openid_config: dict,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.auth.openid_config_url,
        json=openid_config,
    )
    httpx_mock.add_response(
        method="GET",
        url=openid_config["jwks_uri"],
        status_code=404,
    )
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
    ) as client:
        response = await client.get("/extensions", headers={"Authorization": f"Bearer {jwt_token}"})
        assert response.status_code == 401


@pytest.mark.asyncio
async def test_invalid_jwks_data(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    httpx_mock: HTTPXMock,
    settings_factory: SettingsFactory,
    jwt_token: str,
    openid_config: dict,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.auth.openid_config_url,
        json=openid_config,
    )
    httpx_mock.add_response(method="GET", url=openid_config["jwks_uri"], status_code=200, json={})
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
    ) as client:
        response = await client.get("/extensions", headers={"Authorization": f"Bearer {jwt_token}"})
        assert response.status_code == 401


@pytest.mark.asyncio
async def test_jwks_key_not_found(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    httpx_mock: HTTPXMock,
    settings_factory: SettingsFactory,
    jwt_token: str,
    openid_config: dict,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.auth.openid_config_url,
        json=openid_config,
    )
    httpx_mock.add_response(
        method="GET", url=openid_config["jwks_uri"], status_code=200, json={"keys": []}
    )
    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
    ) as client:
        response = await client.get("/extensions", headers={"Authorization": f"Bearer {jwt_token}"})
        assert response.status_code == 401


@pytest.mark.parametrize("exc_type", [jwt.InvalidTokenError, jwt.InvalidKeyError])
async def test_invalid_key_or_token_error(
    api_client: AsyncClient,
    mocker: MockerFixture,
    exc_type: type[Exception],
):
    mocker.patch("mrok.controller.auth.jwt.decode", side_effect=exc_type("bla"))
    resp = await api_client.get("/extensions")
    assert resp.status_code == 401



import shlex

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app
from mrok.cli.commands.admin.unregister.extensions import do_unregister as do_unregister_extension
from mrok.cli.commands.admin.unregister.instances import do_unregister as do_unregister_instance
from tests.conftest import SettingsFactory


def test_unregister_extension_command(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_unregister_coro = mocker.MagicMock()
    mock_unregister = mocker.MagicMock(return_value=mock_unregister_coro)

    mocker.patch("mrok.cli.commands.admin.unregister.extensions.do_unregister", mock_unregister)
    mock_run = mocker.patch("mrok.cli.commands.admin.unregister.extensions.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("admin unregister extension EXT-1234-5678"),
    )
    assert result.exit_code == 0
    mock_run.assert_called_once_with(mock_unregister_coro)

    mock_unregister.assert_called_once_with(
        settings,
        "EXT-1234-5678",
    )


def test_unregister_extension_command_invalid_extension_id(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_unregister_coro = mocker.MagicMock()
    mock_unregister = mocker.MagicMock(return_value=mock_unregister_coro)

    mocker.patch("mrok.cli.commands.admin.unregister.extensions.do_unregister", mock_unregister)
    mocker.patch("mrok.cli.commands.admin.unregister.extensions.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("admin unregister extension EXT-1234"),
    )
    assert result.exit_code == 2
    assert "Invalid value for 'EXTENSION_ID': ext_id must match EXT-xxxx-yyyy" in result.stderr


@pytest.mark.asyncio
async def test_do_unregister_extension(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.__aenter__.return_value = mocked_api
    mocked_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.unregister.extensions.ZitiManagementAPI",
        return_value=mocked_api,
    )
    mocked_unregister_service = mocker.patch(
        "mrok.cli.commands.admin.unregister.extensions.unregister_extension"
    )

    await do_unregister_extension(settings, "EXT-1234")
    mocked_unregister_service.assert_awaited_once_with(
        settings,
        mocked_api,
        "EXT-1234",
    )
    mocked_api_ctor.assert_called_once_with(settings)


# ===========


def test_unregister_instance_command(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_unregister_coro = mocker.MagicMock()
    mock_unregister = mocker.MagicMock(return_value=mock_unregister_coro)

    mocker.patch("mrok.cli.commands.admin.unregister.instances.do_unregister", mock_unregister)
    mock_run = mocker.patch("mrok.cli.commands.admin.unregister.instances.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("admin unregister instance EXT-1234-5678 INS-1234-5678-0001"),
    )
    assert result.exit_code == 0
    mock_run.assert_called_once_with(mock_unregister_coro)

    mock_unregister.assert_called_once_with(
        settings,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
    )


def test_unregister_instance_command_invalid_extension_id(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_unregister_coro = mocker.MagicMock()
    mock_unregister = mocker.MagicMock(return_value=mock_unregister_coro)

    mocker.patch("mrok.cli.commands.admin.unregister.instances.do_unregister", mock_unregister)
    mocker.patch("mrok.cli.commands.admin.unregister.instances.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("admin unregister instance EXT-1234 INS-1234-0001"),
    )
    assert result.exit_code == 2
    assert "Invalid value for 'EXTENSION_ID': ext_id must match EXT-xxxx-yyyy" in result.stderr


@pytest.mark.asyncio
async def test_do_unregister_instance(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.__aenter__.return_value = mocked_api
    mocked_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.unregister.instances.ZitiManagementAPI",
        return_value=mocked_api,
    )
    mocked_unregister_service = mocker.patch(
        "mrok.cli.commands.admin.unregister.instances.unregister_instance"
    )

    await do_unregister_instance(settings, "EXT-1234", "INS-1234-0001")
    mocked_unregister_service.assert_awaited_once_with(
        mocked_api,
        "EXT-1234",
        "INS-1234-0001",
    )
    mocked_api_ctor.assert_called_once_with(settings)



from collections.abc import AsyncGenerator
from typing import Annotated

from fastapi import Depends

from mrok.controller.dependencies.conf import AppSettings
from mrok.ziti import api


class APIClientFactory[T: api.BaseZitiAPI]:
    def __init__(self, client_cls: type[T]):
        self.client_cls = client_cls

    async def __call__(self, settings: AppSettings) -> AsyncGenerator[T]:
        client = self.client_cls(settings)
        async with client:
            yield client


ZitiManagementAPI = Annotated[
    api.ZitiManagementAPI,
    Depends(APIClientFactory(api.ZitiManagementAPI)),
]
ZitiClientAPI = Annotated[
    api.ZitiClientAPI,
    Depends(APIClientFactory(api.ZitiClientAPI)),
]



from fastapi import FastAPI
from pytest_mock import MockerFixture

from mrok.controller.openapi import generate_openapi_spec
from tests.conftest import SettingsFactory


def test_gen_openapi(settings_factory: SettingsFactory, fastapi_app: FastAPI):
    settings = settings_factory()
    assert fastapi_app.openapi_schema is None
    spec = generate_openapi_spec(fastapi_app, settings)
    assert fastapi_app.openapi_schema == spec


def test_gen_openapi_already_generated(
    mocker: MockerFixture, settings_factory: SettingsFactory, fastapi_app: FastAPI
):
    mocked_spec = mocker.MagicMock()
    settings = settings_factory()
    fastapi_app.openapi_schema = mocked_spec
    spec = generate_openapi_spec(fastapi_app, settings)
    assert fastapi_app.openapi_schema == spec



from typing import Annotated, Any

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
)

from mrok.ziti.api import TagsType


class BaseSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True, extra="ignore")
    tags: TagsType | None = None


class IdSchema(BaseModel):
    id: str


class ExtensionIdSchema(BaseModel):
    id: Annotated[str, Field(pattern=r"EXT-\d{4}-\d{4}")]


# For instance
class InstanceIdSchema(BaseModel):
    id: Annotated[str, Field(pattern=r"INS-\d{4}-\d{4}-\d{4}")]


class ExtensionBase(BaseSchema):
    extension: ExtensionIdSchema


class ExtensionRead(BaseSchema, IdSchema):
    name: str

    @computed_field
    def extension(self) -> dict:
        return {"id": self.name.upper()}


class ExtensionCreate(ExtensionBase):
    pass


class InstanceBase(BaseSchema):
    instance: InstanceIdSchema


class InstanceRead(BaseSchema, IdSchema):
    name: str
    identity: dict[str, Any] | None = None

    @computed_field
    def instance(self) -> dict:
        instance_id, _ = self.name.split(".", 1)
        return {"id": instance_id.upper()}

    @computed_field
    def extension(self) -> dict:
        _, extension_id = self.name.split(".", 1)
        return {"id": extension_id.upper()}


class InstanceCreate(InstanceBase):
    pass



import logging
from typing import Any

import jwt

from mrok.ziti import pki
from mrok.ziti.api import TagsType, ZitiClientAPI, ZitiManagementAPI
from mrok.ziti.constants import MROK_SERVICE_TAG_NAME
from mrok.ziti.errors import (
    ProxyIdentityAlreadyExistsError,
    ServiceNotFoundError,
    UserIdentityNotFoundError,
)

logger = logging.getLogger("mrok.ziti")


async def register_instance(
    mgmt_api: ZitiManagementAPI,
    client_api: ZitiClientAPI,
    extension_id: str,
    instance_id: str,
    tags: TagsType | None = None,
):
    service_name = extension_id.lower()
    tags = tags or {}
    tags[MROK_SERVICE_TAG_NAME] = service_name
    service = await mgmt_api.search_service(service_name)
    if not service:
        raise ServiceNotFoundError(f"A service with name `{extension_id}` does not exists.")

    identity_name = f"{instance_id.lower()}.{service_name}"
    service_policy_name = f"{identity_name}:bind"

    identity = await mgmt_api.search_identity(identity_name)
    if identity:
        service_policy = await mgmt_api.search_service_policy(service_policy_name)
        if service_policy:
            await mgmt_api.delete_service_policy(service_policy["id"])
        router_policy = await mgmt_api.search_router_policy(identity_name)
        if router_policy:
            await mgmt_api.delete_router_policy(router_policy["id"])
        await mgmt_api.delete_identity(identity["id"])

    identity_id = await mgmt_api.create_user_identity(identity_name, tags=tags)
    identity = await mgmt_api.get_identity(identity_id)

    identity_json = await _enroll_identity(
        mgmt_api,
        client_api,
        identity_id,
        identity,
        mrok={
            "identity": identity_name,
            "extension": extension_id,
            "instance": instance_id,
        },
    )

    await mgmt_api.create_bind_service_policy(service_policy_name, service["id"], identity_id)
    await mgmt_api.create_router_policy(identity_name, identity_id)

    return identity, identity_json


async def unregister_instance(
    mgmt_api: ZitiManagementAPI,
    extension_id: str,
    instance_id: str,
):
    service_name = extension_id.lower()
    service = await mgmt_api.search_service(service_name)
    if not service:
        raise ServiceNotFoundError(f"A service with name `{extension_id}` does not exists.")

    identity_name = f"{instance_id.lower()}.{service_name}"
    service_policy_name = f"{identity_name}:bind"

    identity = await mgmt_api.search_identity(identity_name)
    if not identity:
        raise UserIdentityNotFoundError(f"Instance `{instance_id}` not found.")

    service_policy = await mgmt_api.search_service_policy(service_policy_name)
    if service_policy:
        await mgmt_api.delete_service_policy(service_policy["id"])
    router_policy = await mgmt_api.search_router_policy(identity_name)
    if router_policy:
        await mgmt_api.delete_router_policy(router_policy["id"])
    await mgmt_api.delete_identity(identity["id"])


async def enroll_proxy_identity(
    mgmt_api: ZitiManagementAPI,
    client_api: ZitiClientAPI,
    identity_name: str,
    tags: TagsType | None = None,
):
    identity = await mgmt_api.search_identity(identity_name)
    if identity:
        raise ProxyIdentityAlreadyExistsError(
            f"A proxy identity with name `{identity_name}` already exists."
        )
    identity_id = await mgmt_api.create_device_identity(identity_name, tags=tags)
    identity_json = await _enroll_identity(mgmt_api, client_api, identity_id)
    logger.info(f"Enrolled proxy identity '{identity_name}'")

    return identity_id, identity_json


async def _enroll_identity(
    mgmt_api: ZitiManagementAPI,
    client_api: ZitiClientAPI,
    identity_id: str,
    identity: dict[str, Any] | None = None,
    mrok: dict[str, str] | None = None,
):
    if identity is None:
        identity = await mgmt_api.get_identity(identity_id)

    claims = _get_enroll_token_claims(identity)
    pkey_pem, csr_pem = pki.generate_key_and_csr(identity_id)

    enroll_response = await client_api.enroll_identity(claims["jti"], csr_pem)
    certificate_pem = enroll_response["data"]["cert"]
    ca_pem = await pki.get_ca_certificates(mgmt_api)

    return _generate_identity_json(
        client_api.base_url,
        pkey_pem,
        certificate_pem,
        ca_pem,
        mrok=mrok,
    )


def _get_enroll_token_claims(identity: dict[str, Any]):
    jwt_token = identity["enrollment"]["ott"]["jwt"]
    return jwt.decode(jwt_token, algorithms=["RS256"], options={"verify_signature": False})


def _generate_identity_json(
    ziti_api_url: str,
    pkey_pem: str,
    certificate_pem: str,
    ca_pem: str,
    mrok: dict | None = None,
) -> dict[str, Any]:
    identity = {
        "ztAPI": ziti_api_url,
        "ztAPIs": None,
        "configTypes": None,
        "id": {
            "key": f"pem:{pkey_pem}",
            "cert": f"pem:{certificate_pem}",
            "ca": f"pem:{ca_pem}",
        },
        "enableHa": False,
    }
    if mrok:
        identity["mrok"] = mrok
    return identity



import typer

from mrok.ziti.api import TagsType


def parse_tags(pairs: list[str] | None) -> TagsType | None:
    if not pairs:
        return None

    result: dict[str, str | bool | None] = {}
    for item in pairs:
        if "=" not in item:
            raise typer.BadParameter(f"Invalid format {item!r}, expected key=value")
        key, raw = item.split("=", 1)
        raw_lower = raw.strip().lower()
        if raw_lower in ("true", "false"):
            val: str | bool | None = raw_lower == "true"
        elif raw == "":
            val = None
        else:
            val = raw
        result[key.strip()] = val
    return result



import abc
import asyncio
import logging
from collections.abc import Awaitable, Callable
from typing import Any

logger = logging.getLogger("mrok.proxy")

Scope = dict[str, Any]
ASGIReceive = Callable[[], Awaitable[dict[str, Any]]]
ASGISend = Callable[[dict[str, Any]], Awaitable[None]]


class BackendNotFoundError(Exception):
    pass


class ForwardAppBase(abc.ABC):
    """Generic HTTP forwarder base class.

    Subclasses must implement `select_backend(scope)` to return an
    (asyncio.StreamReader, asyncio.StreamWriter) pair connected to the
    desired backend. The base class implements the HTTP/1.1 framing
    and streaming logic (requests and responses).
    """

    def __init__(self, read_chunk_size: int = 65536) -> None:
        # number of bytes to read per iteration when streaming bodies
        self._read_chunk_size: int = int(read_chunk_size)

    @abc.abstractmethod
    async def select_backend(
        self,
        scope: Scope,
        headers: dict[str, str],
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter] | tuple[None, None]:
        """Return (reader, writer) connected to the target backend."""

    async def __call__(self, scope: Scope, receive: ASGIReceive, send: ASGISend) -> None:
        """ASGI callable entry point.

        Delegates to smaller helper methods for readability. Subclasses only
        need to implement backend selection.
        """
        if scope.get("type") != "http":
            await send({"type": "http.response.start", "status": 500, "headers": []})
            await send({"type": "http.response.body", "body": b"Unsupported"})
            return

        method = scope.get("method", "GET")
        path_qs = self.format_path(scope)

        headers = list(scope.get("headers", []))
        headers = self.ensure_host_header(headers, scope)
        reader, writer = await self.select_backend(
            scope, {k[0].decode().lower(): k[1].decode() for k in headers}
        )

        if not (reader and writer):
            await send({"type": "http.response.start", "status": 502, "headers": []})
            await send({"type": "http.response.body", "body": b"Bad Gateway"})
            return

        use_chunked = self.ensure_chunked_if_needed(headers)

        await self.write_request_line_and_headers(writer, method, path_qs, headers)

        await self.stream_request_body(receive, writer, use_chunked)

        status_line = await reader.readline()
        if not status_line:
            await send({"type": "http.response.start", "status": 502, "headers": []})
            await send({"type": "http.response.body", "body": b"Bad Gateway"})
            writer.close()
            await writer.wait_closed()
            return

        status, headers_out, raw_headers = await self.read_status_and_headers(reader, status_line)

        await send({"type": "http.response.start", "status": status, "headers": headers_out})

        await self.stream_response_body(reader, send, raw_headers)

        writer.close()
        await writer.wait_closed()

    def format_path(self, scope: Scope) -> str:
        raw_path = scope.get("raw_path")
        if raw_path:
            return raw_path.decode()
        q = scope.get("query_string", b"")
        path = scope.get("path", "/")
        path_qs = path
        if q:
            path_qs += "?" + q.decode()
        return path_qs

    def ensure_host_header(
        self, headers: list[tuple[bytes, bytes]], scope: Scope
    ) -> list[tuple[bytes, bytes]]:
        if any(n.lower() == b"host" for n, _ in headers):
            return headers
        server = scope.get("server")
        if server:
            host = f"{server[0]}:{server[1]}" if server[1] else server[0]
            headers.append((b"host", host.encode()))
        return headers

    def ensure_chunked_if_needed(self, headers: list[tuple[bytes, bytes]]) -> bool:
        has_content_length = any(n.lower() == b"content-length" for n, _ in headers)
        has_transfer_encoding = any(n.lower() == b"transfer-encoding" for n, _ in headers)
        if not has_content_length and not has_transfer_encoding:
            headers.append((b"transfer-encoding", b"chunked"))
            return True
        return False

    async def write_request_line_and_headers(
        self,
        writer: asyncio.StreamWriter,
        method: str,
        path_qs: str,
        headers: list[tuple[bytes, bytes]],
    ) -> None:
        writer.write(f"{method} {path_qs} HTTP/1.1\r\n".encode())
        for name, value in headers:
            if name.lower() == b"expect":
                continue
            writer.write(name + b": " + value + b"\r\n")
        writer.write(b"\r\n")
        await writer.drain()

    async def stream_request_body(
        self, receive: ASGIReceive, writer: asyncio.StreamWriter, use_chunked: bool
    ) -> None:
        if use_chunked:
            await self.stream_request_chunked(receive, writer)
            return

        await self.stream_request_until_end(receive, writer)

    async def stream_request_chunked(
        self, receive: ASGIReceive, writer: asyncio.StreamWriter
    ) -> None:
        """Send request body to backend using HTTP/1.1 chunked encoding."""
        while True:
            event = await receive()
            if event["type"] == "http.request":
                body = event.get("body", b"") or b""
                if body:
                    writer.write(f"{len(body):X}\r\n".encode())
                    writer.write(body)
                    writer.write(b"\r\n")
                    await writer.drain()
                if not event.get("more_body", False):
                    break
            elif event["type"] == "http.disconnect":
                writer.close()
                return

        writer.write(b"0\r\n\r\n")
        await writer.drain()

    async def stream_request_until_end(
        self, receive: ASGIReceive, writer: asyncio.StreamWriter
    ) -> None:
        """Send request body to backend when content length/transfer-encoding
        already provided (no chunking).
        """
        while True:
            event = await receive()
            if event["type"] == "http.request":
                body = event.get("body", b"") or b""
                if body:
                    writer.write(body)
                    await writer.drain()
                if not event.get("more_body", False):
                    break
            elif event["type"] == "http.disconnect":
                writer.close()
                return

    async def read_status_and_headers(
        self, reader: asyncio.StreamReader, first_line: bytes
    ) -> tuple[int, list[tuple[bytes, bytes]], dict[bytes, bytes]]:
        parts = first_line.decode(errors="ignore").split(" ", 2)
        status = int(parts[1]) if len(parts) >= 2 and parts[1].isdigit() else 502
        headers: list[tuple[bytes, bytes]] = []
        raw_headers: dict[bytes, bytes] = {}
        while True:
            line = await reader.readline()
            if line in (b"\r\n", b"\n", b""):
                break
            i = line.find(b":")
            if i == -1:
                continue
            name = line[:i].strip().lower()
            value = line[i + 1 :].strip()
            headers.append((name, value))
            raw_headers[name] = value

        return status, headers, raw_headers

    def is_chunked(self, te_value: bytes) -> bool:
        """Return True if transfer-encoding header tokens include 'chunked'."""
        if not te_value:
            return False
        # split on commas, strip spaces and check tokens
        tokens = [t.strip() for t in te_value.split(b",")]
        return any(t.lower() == b"chunked" for t in tokens)

    def parse_content_length(self, cl_value: bytes | None) -> int | None:
        """Parse Content-Length header value to int, or return None if invalid."""
        if cl_value is None:
            return None
        try:
            return int(cl_value)
        except Exception:
            return None

    async def drain_trailers(self, reader: asyncio.StreamReader) -> None:
        """Consume trailer header lines until an empty line is encountered."""
        while True:
            trailer = await reader.readline()
            if trailer in (b"\r\n", b"\n", b""):
                break

    async def stream_response_chunked(self, reader: asyncio.StreamReader, send: ASGISend) -> None:
        """Read chunked-encoded response from reader, decode and forward to ASGI send."""
        while True:
            size_line = await reader.readline()
            if not size_line:
                break
            size_str = size_line.split(b";", 1)[0].strip()
            try:
                size = int(size_str, 16)
            except Exception:
                break
            if size == 0:
                # consume trailers
                await self.drain_trailers(reader)
                break
            try:
                chunk = await reader.readexactly(size)
            except Exception:
                break
            # consume the CRLF after the chunk
            try:
                await reader.readexactly(2)
            except Exception:
                logger.warning("failed to read CRLF after chunk from backend")
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def stream_response_with_content_length(
        self, reader: asyncio.StreamReader, send: ASGISend, content_length: int
    ) -> None:
        """Read exactly content_length bytes and forward to ASGI send events."""
        remaining = content_length
        sent_final = False
        while remaining > 0:
            to_read = min(self._read_chunk_size, remaining)
            chunk = await reader.read(to_read)
            if not chunk:
                break
            remaining -= len(chunk)
            more = remaining > 0
            await send({"type": "http.response.body", "body": chunk, "more_body": more})
            if not more:
                sent_final = True

        if not sent_final:
            await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def stream_response_until_eof(self, reader: asyncio.StreamReader, send: ASGISend) -> None:
        """Read from reader until EOF and forward chunks to ASGI send events."""
        while True:
            chunk = await reader.read(self._read_chunk_size)
            if not chunk:
                break
            await send({"type": "http.response.body", "body": chunk, "more_body": True})
        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def stream_response_body(
        self, reader: asyncio.StreamReader, send: ASGISend, raw_headers: dict[bytes, bytes]
    ) -> None:
        te = raw_headers.get(b"transfer-encoding", b"").lower()
        cl = raw_headers.get(b"content-length")

        if self.is_chunked(te):
            await self.stream_response_chunked(reader, send)
            return

        content_length = self.parse_content_length(cl)
        if content_length is not None:
            await self.stream_response_with_content_length(reader, send, content_length)
            return

        await self.stream_response_until_eof(reader, send)



import multiprocessing
from pathlib import Path
from typing import Annotated

import typer

from mrok.agent import sidecar


def number_of_workers():
    return (multiprocessing.cpu_count() * 2) + 1


default_workers = number_of_workers()


def register(app: typer.Typer) -> None:
    @app.command("sidecar")
    def run_sidecar(
        identity_file: Path = typer.Argument(
            ...,
            help="Identity json file",
        ),
        target: Path = typer.Argument(
            ...,
            help="Target service (host:port or path to unix domain socket)",
        ),
        workers: Annotated[
            int,
            typer.Option(
                "--workers",
                "-w",
                help=f"Number of workers. Default: {default_workers}",
                show_default=True,
            ),
        ] = default_workers,
        reload: Annotated[
            bool,
            typer.Option(
                "--reload",
                "-r",
                help="Enable auto-reload. Default: False",
                show_default=True,
            ),
        ] = False,
    ):
        """Run a Sidecar Proxy to expose a web application through OpenZiti."""
        if ":" in str(target):
            host, port = str(target).split(":", 1)
            target_addr = (host or "127.0.0.1", int(port))
        else:
            target_addr = str(target)  # type: ignore

        sidecar.run(str(identity_file), target_addr, workers=workers, reload=reload)



import pytest
from pytest_mock import MockerFixture

from mrok.http.config import MrokBackendConfig
from mrok.http.lifespan import MrokLifespan
from mrok.http.protocol import MrokHttpToolsProtocol


def test_backend_config_init(ziti_identity_file: str):
    async def fake_asgi_app(scope, receive, send):
        pass

    config = MrokBackendConfig(fake_asgi_app, ziti_identity_file)
    assert config.app == fake_asgi_app
    assert config.loop == "asyncio"
    assert config.http == MrokHttpToolsProtocol
    assert config.lifespan == "auto"
    assert config.backlog == 2048
    assert config.service_name == "ext-1234-5678"
    assert config.identity_name == "ins-1234-5678-0001.ext-1234-5678"
    assert config.identity_file == ziti_identity_file


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_backend_config_load(ziti_identity_file: str):
    async def fake_asgi_app(scope, receive, send):
        pass

    config = MrokBackendConfig(fake_asgi_app, ziti_identity_file)
    config.load()
    assert config.app == fake_asgi_app
    assert config.loop == "asyncio"
    assert config.http_protocol_class == MrokHttpToolsProtocol
    assert config.lifespan_class == MrokLifespan


def test_backend_config_bind_socket(mocker: MockerFixture, ziti_identity_file: str):
    mocked_ctx = mocker.MagicMock()
    mocked_socket = mocker.MagicMock()
    mocked_ctx.bind.return_value = mocked_socket
    mocked_openziti_load = mocker.patch(
        "mrok.http.config.openziti.load",
        return_value=(mocked_ctx, 0),
    )

    async def fake_asgi_app(scope, receive, send):
        pass

    config = MrokBackendConfig(
        fake_asgi_app, ziti_identity_file, backlog=4096, ziti_load_timeout_ms=1234
    )
    assert config.bind_socket() == mocked_socket
    mocked_openziti_load.assert_called_once_with(ziti_identity_file, timeout=1234)
    mocked_ctx.bind.assert_called_once_with(
        "ext-1234-5678",
    )
    mocked_socket.listen.assert_called_once_with(4096)


def test_backend_config_bind_socket_ziti_error(mocker: MockerFixture, ziti_identity_file: str):
    mocked_ctx = mocker.MagicMock()
    mocked_socket = mocker.MagicMock()
    mocked_ctx.bind.return_value = mocked_socket
    mocker.patch(
        "mrok.http.config.openziti.load",
        return_value=(None, 1),
    )

    async def fake_asgi_app(scope, receive, send):
        pass

    config = MrokBackendConfig(fake_asgi_app, ziti_identity_file, backlog=4096)
    with pytest.raises(RuntimeError) as cv:
        config.bind_socket()

    assert str(cv.value) == f"Failed to load Ziti identity from {ziti_identity_file}: 1"


def test_backend_config_configure_logging(mocker: MockerFixture, ziti_identity_file: str):
    mocked_settings = mocker.MagicMock()
    mocker.patch("mrok.http.config.get_settings", return_value=mocked_settings)
    mocked_setup_logging = mocker.patch("mrok.http.config.setup_logging")

    async def fake_asgi_app(scope, receive, send):
        pass

    MrokBackendConfig(fake_asgi_app, ziti_identity_file, backlog=4096)
    mocked_setup_logging.assert_called_once_with(mocked_settings)



import json
import logging
import socket
from collections.abc import Callable
from pathlib import Path
from typing import Any

import openziti
from uvicorn import config

from mrok.conf import get_settings
from mrok.http.protocol import MrokHttpToolsProtocol
from mrok.logging import setup_logging

logger = logging.getLogger("mrok.proxy")

config.LIFESPAN["auto"] = "mrok.http.lifespan:MrokLifespan"

ASGIApplication = config.ASGIApplication


class MrokBackendConfig(config.Config):
    def __init__(
        self,
        app: ASGIApplication | Callable[..., Any] | str,
        identity_file: str | Path,
        ziti_load_timeout_ms: int = 5000,
        backlog: int = 2048,
    ):
        self.identity_file = identity_file
        self.ziti_load_timeout_ms = ziti_load_timeout_ms
        self.service_name, self.identity_name, self.instance_id = self.get_identity_info(
            identity_file
        )
        super().__init__(
            app,
            loop="asyncio",
            http=MrokHttpToolsProtocol,
            backlog=backlog,
        )

    def get_identity_info(self, identity_file: str | Path):
        with open(identity_file) as f:
            identity_data = json.load(f)
            try:
                identity_name = identity_data["mrok"]["identity"]
                instance_id, service_name = identity_name.split(".", 1)
                return service_name, identity_name, instance_id
            except KeyError:
                raise ValueError("Invalid identity file: identity file is not mrok compatible.")

    def bind_socket(self) -> socket.socket:
        logger.info(f"Connect to Ziti service '{self.service_name} ({self.instance_id})'")

        ctx, err = openziti.load(str(self.identity_file), timeout=self.ziti_load_timeout_ms)
        if err != 0:
            raise RuntimeError(f"Failed to load Ziti identity from {self.identity_file}: {err}")

        sock = ctx.bind(self.service_name)
        sock.listen(self.backlog)
        logger.info(f"listening on ziti service {self.service_name} for connections")
        return sock

    def configure_logging(self) -> None:
        setup_logging(get_settings())



import base64

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization.pkcs7 import load_der_pkcs7_certificates
from cryptography.x509.oid import NameOID

from mrok.ziti.api import ZitiManagementAPI

_ca_certificates = None


async def get_ca_certificates(mgmt_api: ZitiManagementAPI) -> str:
    global _ca_certificates
    if not _ca_certificates:
        cas_pkcs7 = await mgmt_api.fetch_ca_certificates()
        pkcs7_bytes = base64.b64decode(cas_pkcs7)
        pkcs7_certs = load_der_pkcs7_certificates(pkcs7_bytes)
        ca_certificates = []
        for cert in pkcs7_certs:
            cert_pem = cert.public_bytes(serialization.Encoding.PEM)
            ca_certificates.append(cert_pem.decode("utf-8"))
        _ca_certificates = "\n".join(ca_certificates)
    return _ca_certificates


def generate_key_and_csr(identity_id: str, key_size: int = 4096) -> tuple[str, str]:
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "CH"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "SoftwareOne"),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "Marketplace Platform"),
            x509.NameAttribute(NameOID.COMMON_NAME, identity_id),
        ]
    )
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(subject)
        .sign(private_key, hashes.SHA256())
    )

    key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode()

    csr_pem = csr.public_bytes(serialization.Encoding.PEM).decode()

    return key_pem, csr_pem



from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi

from mrok.conf import Settings


def generate_openapi_spec(app: FastAPI, settings: Settings):
    if app.openapi_schema:  # pragma: no cover
        return app.openapi_schema

    # for api_route in app.routes:
    #     if isinstance(api_route, APIRoute):
    #         for dep in api_route.dependant.dependencies:
    #             if dep.call and isinstance(dep.call, RQLQuery):
    #                 api_route.description = (
    #                     f"{api_route.description}\n\n"
    #                     "## Available RQL filters\n\n"
    #                     f"{dep.call.rules.get_documentation()}"
    #                 )

    spec = get_openapi(
        title=app.title,
        version=app.version,
        openapi_version=app.openapi_version,
        description=app.description,
        tags=app.openapi_tags,
        routes=app.routes,
    )
    # spec = inject_code_samples(
    #     spec,
    #     SnippetRenderer(),
    #     settings.api_base_url,
    # )
    app.openapi_schema = spec
    return app.openapi_schema



from urllib.parse import quote

import pytest
from httpx import AsyncClient
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from mrok.ziti.api import ZitiClientAPI, ZitiManagementAPI
from mrok.ziti.constants import MROK_SERVICE_TAG_NAME, MROK_VERSION_TAG_NAME
from tests.conftest import SettingsFactory


@pytest.mark.asyncio
async def test_list_instances(
    mocker: MockerFixture,
    api_client: AsyncClient,
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    mocker.patch(
        "mrok.controller.routes.fetch_extension_or_404", return_value={"name": "ext-1234-5678"}
    )
    query = quote(f'tags.{MROK_SERVICE_TAG_NAME} = "ext-1234-5678"')
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/identities?filter={query}&limit=5&offset=0",
        json={
            "meta": {"pagination": {"totalCount": 10, "limit": 5, "offset": 0}},
            "data": [{"id": f"ins{i}", "name": "ins.svc"} for i in range(5)],
        },
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/identities?filter={query}&limit=5&offset=5",
        json={
            "meta": {"pagination": {"totalCount": 10, "limit": 5, "offset": 5}},
            "data": [{"id": f"ins{5 + i}", "name": "ins.svc"} for i in range(5)],
        },
    )
    response = await api_client.get("/extensions/ext-1234-5678/instances?limit=5&offset=0")
    assert response.status_code == 200
    page_1 = response.json()
    assert len(page_1["data"]) == 5
    assert page_1["$meta"]["pagination"]["offset"] == 0

    response = await api_client.get("/extensions/ext-1234-5678/instances?limit=5&offset=5")
    assert response.status_code == 200
    page_2 = response.json()
    assert len(page_2["data"]) == 5
    assert page_2["$meta"]["pagination"]["offset"] == 5

    assert sorted(page_1["data"], key=lambda x: x["id"]) != sorted(
        page_2["data"], key=lambda x: x["id"]
    )


@pytest.mark.asyncio
async def test_register_instance(mocker: MockerFixture, api_client: AsyncClient):
    mocker.patch(
        "mrok.controller.routes.fetch_extension_or_404", return_value={"name": "ext-1234-5678"}
    )
    mocked_register = mocker.patch(
        "mrok.controller.routes.register_instance",
        return_value=(
            {
                "id": "a1b2cd",
                "name": "ins-1234-5678-0001.ext-1234-5678",
                "tags": {
                    MROK_VERSION_TAG_NAME: "0.0.0.dev0",
                    MROK_SERVICE_TAG_NAME: "ext-1234-5678",
                    "account": "ACC-1234-5678",
                },
            },
            {
                "identity": "json",
            },
        ),
    )

    response = await api_client.post(
        "/extensions/ext-1234-5678/instances",
        json={
            "instance": {"id": "INS-1234-5678-0001"},
            "tags": {"account": "ACC-1234-5678"},
        },
    )
    assert response.status_code == 201
    assert response.json() == {
        "id": "a1b2cd",
        "extension": {"id": "EXT-1234-5678"},
        "instance": {"id": "INS-1234-5678-0001"},
        "name": "ins-1234-5678-0001.ext-1234-5678",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
            MROK_SERVICE_TAG_NAME: "ext-1234-5678",
            "account": "ACC-1234-5678",
        },
        "identity": {"identity": "json"},
    }
    assert mocked_register.call_count == 1
    assert isinstance(mocked_register.mock_calls[0].args[0], ZitiManagementAPI)
    assert isinstance(mocked_register.mock_calls[0].args[1], ZitiClientAPI)
    assert mocked_register.mock_calls[0].args[2] == "ext-1234-5678"
    assert mocked_register.mock_calls[0].args[3] == "INS-1234-5678-0001"
    assert mocked_register.mock_calls[0].args[4] == {"account": "ACC-1234-5678"}


@pytest.mark.asyncio
async def test_get_instance(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    api_client: AsyncClient,
    httpx_mock: HTTPXMock,
):
    mocker.patch(
        "mrok.controller.routes.fetch_extension_or_404", return_value={"name": "ext-1234-5678"}
    )
    settings = settings_factory()
    query = quote(
        '(id="INS-1234-5678-0001.ext-1234-5678" or name="ins-1234-5678-0001.ext-1234-5678") '
        f"and tags.{MROK_VERSION_TAG_NAME} != null"
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/identities?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 1}},
            "data": [
                {
                    "id": "ins1",
                    "name": "ins-1234-5678-0001.ext-1234-5678",
                    "tags": {
                        MROK_VERSION_TAG_NAME: "0.0.0.dev0",
                        MROK_SERVICE_TAG_NAME: "ext-1234-5678",
                    },
                }
            ],
        },
    )

    response = await api_client.get("/extensions/EXT-1234-5678/instances/INS-1234-5678-0001")
    assert response.status_code == 200
    assert response.json() == {
        "id": "ins1",
        "extension": {"id": "EXT-1234-5678"},
        "instance": {"id": "INS-1234-5678-0001"},
        "name": "ins-1234-5678-0001.ext-1234-5678",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
            MROK_SERVICE_TAG_NAME: "ext-1234-5678",
        },
    }


@pytest.mark.asyncio
async def test_get_instance_by_instance_id(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    api_client: AsyncClient,
    httpx_mock: HTTPXMock,
):
    mocker.patch(
        "mrok.controller.routes.fetch_extension_or_404", return_value={"name": "ext-1234-5678"}
    )
    settings = settings_factory()
    query = quote(f'(id="ins1" or name="ins1") and tags.{MROK_VERSION_TAG_NAME} != null')
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/identities?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 1}},
            "data": [
                {
                    "id": "ins1",
                    "name": "ins-1234-5678-0001.ext-1234-5678",
                    "tags": {
                        MROK_VERSION_TAG_NAME: "0.0.0.dev0",
                        MROK_SERVICE_TAG_NAME: "ext-1234-5678",
                    },
                }
            ],
        },
    )

    response = await api_client.get("/extensions/EXT-1234-5678/instances/ins1")
    assert response.status_code == 200
    assert response.json() == {
        "id": "ins1",
        "extension": {"id": "EXT-1234-5678"},
        "instance": {"id": "INS-1234-5678-0001"},
        "name": "ins-1234-5678-0001.ext-1234-5678",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
            MROK_SERVICE_TAG_NAME: "ext-1234-5678",
        },
    }


@pytest.mark.asyncio
async def test_get_instance_not_found(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    api_client: AsyncClient,
    httpx_mock: HTTPXMock,
):
    mocker.patch(
        "mrok.controller.routes.fetch_extension_or_404", return_value={"name": "ext-1234-5678"}
    )
    settings = settings_factory()
    query = quote(
        '(id="INS-1234-5678-0001.ext-1234-5678" or name="ins-1234-5678-0001.ext-1234-5678") '
        f"and tags.{MROK_VERSION_TAG_NAME} != null"
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/identities?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 10}},
            "data": [],
        },
    )

    response = await api_client.get("/extensions/EXT-1234-5678/instances/INS-1234-5678-0001")
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_delete_instance(
    mocker: MockerFixture,
    api_client: AsyncClient,
):
    mocker.patch(
        "mrok.controller.routes.fetch_instance_or_404",
        return_value={"name": "ins-1234-5678.ext-1234-5678"},
    )
    mocker.patch(
        "mrok.controller.routes.unregister_instance",
    )

    response = await api_client.delete("/extensions/EXT-1234-5678/instances/INS-1234-5678-0001")
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_delete_extension_not_found(
    mocker: MockerFixture,
    api_client: AsyncClient,
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    mocker.patch(
        "mrok.controller.routes.fetch_extension_or_404", return_value={"name": "ext-1234-5678"}
    )

    query = quote(
        '(id="INS-1234-5678-0001.ext-1234-5678" or name="ins-1234-5678-0001.ext-1234-5678") '
        f"and tags.{MROK_VERSION_TAG_NAME} != null"
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/identities?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 0}},
            "data": [],
        },
    )

    response = await api_client.delete("/extensions/EXT-1234-5678/instances/INS-1234-5678-0001")
    assert response.status_code == 404



import pytest
from pytest_mock import MockerFixture

from mrok.http.config import MrokBackendConfig
from mrok.http.server import MrokServer


@pytest.mark.asyncio
async def test_serve(mocker: MockerFixture):
    mocked_socket = mocker.MagicMock()
    mocker.patch.object(MrokBackendConfig, "bind_socket", return_value=mocked_socket)
    mocker.patch.object(
        MrokBackendConfig, "get_identity_info", return_value=("ext", "ins.ext", "ins")
    )
    mocked_inner_serve = mocker.patch.object(MrokServer, "_serve")

    async def fake_asgi_app(scope, receive, send):
        pass

    config = MrokBackendConfig(fake_asgi_app, "ziti-identity.json")

    server = MrokServer(config)
    await server.serve()
    mocked_inner_serve.assert_awaited_once_with([mocked_socket])


@pytest.mark.asyncio
async def test_serve_with_socket(mocker: MockerFixture):
    mocked_socket = mocker.MagicMock()
    mocked_bind = mocker.patch.object(MrokBackendConfig, "bind_socket")
    mocker.patch.object(
        MrokBackendConfig, "get_identity_info", return_value=("ext", "ins.ext", "ins")
    )
    mocked_inner_serve = mocker.patch.object(MrokServer, "_serve")

    async def fake_asgi_app(scope, receive, send):
        pass

    config = MrokBackendConfig(fake_asgi_app, "ziti-identity.json")

    server = MrokServer(config)
    await server.serve([mocked_socket])
    mocked_inner_serve.assert_awaited_once_with([mocked_socket])
    mocked_bind.assert_not_called()



import logging
import os
import signal
import threading
from collections.abc import Callable
from pathlib import Path

from watchfiles import watch
from watchfiles.filters import PythonFilter
from watchfiles.run import CombinedProcess, start_process

logger = logging.getLogger("mrok.agent")


def print_path(path):
    try:
        return f'"{path.relative_to(Path.cwd())}"'
    except ValueError:
        return f'"{path}"'


class Master:
    def __init__(
        self,
        start_fn: Callable,
        workers: int,
        reload: bool,
    ):
        self.start_fn = start_fn
        self.workers = workers
        self.reload = reload
        self.worker_processes: list[CombinedProcess] = []
        self.stop_event = threading.Event()
        self.watch_filter = PythonFilter(ignore_paths=None)
        self.watcher = watch(
            Path.cwd(),
            watch_filter=self.watch_filter,
            stop_event=self.stop_event,
            yield_on_timeout=True,
        )
        self.setup_signals_handler()

    def setup_signals_handler(self):
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, self.handle_signal)

    def handle_signal(self, *args, **kwargs):
        self.stop_event.set()

    def start(self):
        for _ in range(self.workers):
            p = start_process(
                self.start_fn,
                "function",
                (),
                None,
            )
            logger.info(f"Worker [{p.pid}] started")
            self.worker_processes.append(p)

    def stop(self):
        for process in self.worker_processes:
            process.stop(sigint_timeout=5, sigkill_timeout=1)
        self.worker_processes = []

    def restart(self):
        self.stop()
        self.start()

    def __iter__(self):
        return self

    def __next__(self):
        changes = next(self.watcher)
        if changes:
            return list({Path(change[1]) for change in changes})
        return None

    def run(self):
        self.start()
        logger.info(f"Master process started: {os.getpid()}")
        if self.reload:
            for files_changed in self:
                if files_changed:
                    logger.warning(
                        f"{', '.join(map(print_path, files_changed))} changed, reloading...",
                    )
                    self.restart()
        else:
            self.stop_event.wait()



from dynaconf import Dynaconf, LazySettings

type Settings = LazySettings

DEFAULT_SETTINGS = {
    "LOGGING": {
        "debug": False,
        "rich": False,
    },
    "PROXY": {
        "identity": "public",
        "mode": "zrok",
    },
    "ZITI": {
        "ssl_verify": False,
    },
    "PAGINATION": {"limit": 50},
}

_settings = None


def get_settings() -> Settings:
    global _settings
    if not _settings:
        _settings = Dynaconf(
            envvar_prefix="MROK",
            settings_files=["settings.yaml", ".secrets.yaml"],
            merge_enabled=True,
        )
        _settings.configure(**DEFAULT_SETTINGS)
    return _settings



import logging
from typing import Any

from mrok.conf import Settings
from mrok.ziti.api import TagsType, ZitiManagementAPI
from mrok.ziti.errors import (
    ConfigTypeNotFoundError,
    ProxyIdentityNotFoundError,
    ServiceAlreadyRegisteredError,
    ServiceNotFoundError,
)

logger = logging.getLogger(__name__)


async def register_extension(
    settings: Settings, mgmt_api: ZitiManagementAPI, extension_id: str, tags: TagsType | None
) -> dict[str, Any]:
    service_name = extension_id.lower()
    registered = False
    proxy_identity = await mgmt_api.search_identity(settings.proxy.identity)
    if not proxy_identity:
        raise ProxyIdentityNotFoundError(
            f"Identity for proxy `{settings.proxy.identity}` not found.",
        )

    config_type = await mgmt_api.search_config_type(f"{settings.proxy.mode}.proxy.v1")
    if not config_type:
        raise ConfigTypeNotFoundError(f"Config type `{settings.proxy.mode}.proxy.v1` not found.")

    config = await mgmt_api.search_config(service_name)
    if not config:
        config_id = await mgmt_api.create_config(service_name, config_type["id"], tags=tags)
        registered = True
    else:
        config_id = config["id"]
    service = await mgmt_api.search_service(service_name)
    if not service:
        service_id = await mgmt_api.create_service(service_name, config_id, tags=tags)
        service = await mgmt_api.get_service(service_id)
        registered = True
    else:
        service_id = service["id"]
    proxy_identity_id = proxy_identity["id"]
    service_policy_name = f"{service_name}:{settings.proxy.identity}:dial"
    dial_service_policy = await mgmt_api.search_service_policy(service_policy_name)
    if not dial_service_policy:
        await mgmt_api.create_dial_service_policy(
            service_policy_name,
            service_id,
            proxy_identity_id,
            tags=tags,
        )
        registered = True

    router_policy = await mgmt_api.search_service_router_policy(service_name)
    if not router_policy:
        await mgmt_api.create_service_router_policy(service_name, service_id, tags=tags)
        registered = True
    if not registered:
        raise ServiceAlreadyRegisteredError(f"Extension `{extension_id}` already registered.")
    return service


async def unregister_extension(
    settings: Settings, mgmt_api: ZitiManagementAPI, extension_id: str
) -> None:
    service_name = extension_id.lower()
    service = await mgmt_api.search_service(service_name)
    if not service:
        raise ServiceNotFoundError(f"Extension `{extension_id}` not found.")

    router_policy = await mgmt_api.search_service_router_policy(service_name)
    if router_policy:
        await mgmt_api.delete_service_router_policy(router_policy["id"])

    service_policy_name = f"{service_name}:{settings.proxy.identity}:dial"

    dial_service_policy = await mgmt_api.search_service_policy(service_policy_name)
    if dial_service_policy:
        await mgmt_api.delete_service_policy(dial_service_policy["id"])

    config = await mgmt_api.search_config(service_name)
    if config:
        await mgmt_api.delete_config(config["id"])

    await mgmt_api.delete_service(service["id"])



from mrok.ziti.api import (
    ZitiAPIError,
    ZitiAuthError,
    ZitiBadRequestError,
    ZitiClientAPI,
    ZitiManagementAPI,
)

__all__ = [
    "ZitiAPIError",
    "ZitiAuthError",
    "ZitiBadRequestError",
    "ZitiClientAPI",
    "ZitiManagementAPI",
]



import asyncio
import contextlib
import os
from functools import partial

from mrok.http.config import ASGIApplication, MrokBackendConfig
from mrok.http.master import Master
from mrok.http.server import MrokServer


def run_ziticorn(app: ASGIApplication | str, identity_file: str):
    import sys

    sys.path.insert(0, os.getcwd())
    config = MrokBackendConfig(app, identity_file)
    server = MrokServer(config)
    with contextlib.suppress(KeyboardInterrupt, asyncio.CancelledError):
        server.run()


def run(
    app: ASGIApplication | str,
    identity_file: str,
    workers: int = 4,
    reload: bool = False,
):
    start_fn = partial(run_ziticorn, app, identity_file)
    master = Master(start_fn, workers=workers, reload=reload)
    master.run()



import json
import logging
import ssl
import tempfile
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator
from functools import cached_property
from types import TracebackType
from typing import Any, Literal

import httpx

from mrok.conf import Settings
from mrok.ziti.constants import MROK_VERSION_TAG, MROK_VERSION_TAG_NAME

logger = logging.getLogger(__name__)

TagsType = dict[str, str | bool | None]


class ZitiAPIError(Exception):
    pass


class ZitiAuthError(ZitiAPIError):
    pass


class ZitiBadRequestError(ZitiAPIError):
    def __init__(self, response: dict[str, Any]):
        self.response = response

    def __str__(self) -> str:
        err = self.response["error"]
        cause = err["cause"]
        return f"{err['code']} - {err['message']} ({cause['field']}: {cause['reason']})"


class BaseZitiAPI(ABC):
    def __init__(self, settings: Settings):
        self.settings = settings
        self.limit = self.settings.pagination.limit
        self.token = None

    @property
    @abstractmethod
    def base_url(self):
        raise NotImplementedError("base_url property must be implemented in subclasses")

    @property
    def auth(self):
        if self.settings.ziti.auth.get("username") and self.settings.ziti.auth.get("password"):
            return ZitiPasswordAuth(self)
        elif self.settings.ziti.auth.get("identity"):
            return ZitiIdentityAuth(self)
        else:
            raise ZitiAuthError("Unsupported authentication method for OpenZiti.")

    @cached_property
    def httpx_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(
            base_url=self.base_url,
            auth=self.auth,
            verify=self.settings.ziti.ssl_verify,
            timeout=httpx.Timeout(
                connect=0.25,
                read=self.settings.ziti.read_timeout,
                write=2.0,
                pool=5.0,
            ),
        )

    async def create(self, endpoint: str, payload: dict[str, Any], tags: TagsType | None) -> str:
        payload["tags"] = self._merge_tags(tags)
        response: httpx.Response = await self.httpx_client.post(
            endpoint,
            json=payload,
        )
        if response.status_code == 400:
            raise ZitiBadRequestError(response.json())
        response.raise_for_status()
        return response.json()["data"]["id"]

    async def get(self, endpoint: str, id: str) -> dict[str, Any]:
        response = await self.httpx_client.get(f"{endpoint}/{id}")
        response.raise_for_status()
        return response.json()["data"]

    async def delete(self, endpoint: str, id: str) -> None:
        response = await self.httpx_client.delete(f"{endpoint}/{id}")
        response.raise_for_status()
        return response.json()

    async def search_by_id_or_name(self, endpoint: str, id_or_name: str) -> dict[str, Any] | None:
        query = (
            f'(id="{id_or_name}" or name="{id_or_name.lower()}") '
            f"and tags.{MROK_VERSION_TAG_NAME} != null"
        )
        response = await self.httpx_client.get(
            endpoint,
            params={"filter": query},
        )
        if response.status_code == 400:
            raise ZitiBadRequestError(response.json())
        response.raise_for_status()
        response_data = response.json()
        if response_data["meta"]["pagination"]["totalCount"] == 1:
            return response_data["data"][0]

    async def get_page(
        self, endpoint: str, limit: int, offset: int, params: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        params = params or {}
        params["limit"] = limit
        params["offset"] = offset
        page_response = await self.httpx_client.get(endpoint, params=params)
        page_response.raise_for_status()
        page = page_response.json()
        return page

    async def collection_iterator(
        self, endpoint: str, params: dict[str, Any] | None = None
    ) -> AsyncGenerator[dict, None]:
        offset = 0
        while True:
            page = await self.get_page(endpoint, self.limit, offset, params=params)
            items = page["data"]

            for item in items:
                yield item

            pagination_meta = page["meta"]["pagination"]
            total = pagination_meta["totalCount"]
            if total <= self.limit + offset:
                break

            offset = offset + self.limit

    async def __aenter__(self):
        await self.httpx_client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        return await self.httpx_client.__aexit__(exc_type, exc_val, exc_tb)

    def _merge_tags(self, tags: TagsType | None) -> TagsType:
        prepared_tags: TagsType = tags or {}
        prepared_tags.update(MROK_VERSION_TAG)
        return prepared_tags


class BaseZitiAuth(httpx.Auth):
    def __init__(self, api: BaseZitiAPI):
        self.api = api

    def update_token(self, response: httpx.Response) -> None:
        response.raise_for_status()
        data = response.json()
        self.api.token = data["data"]["token"]


class ZitiIdentityAuthContext:
    def __init__(self, identity_path: str):
        identity = json.load(open(identity_path))
        cert_data = identity["id"]["cert"][4:]
        key_data = identity["id"]["key"][4:]
        ca_data = identity["id"]["ca"][4:]
        with (
            tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".pem") as key_file,
            tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".pem") as cert_file,
            tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".pem") as ca_file,
        ):
            key_file.write(key_data)
            key_file.flush()
            cert_file.write(cert_data)
            cert_file.flush()
            ca_file.write(ca_data)
            ca_file.flush()
            self.ssl_context = ssl.create_default_context(
                cafile=ca_file.name,
            )
            self.ssl_context.load_cert_chain(
                keyfile=key_file.name,
                certfile=cert_file.name,
            )


class ZitiPasswordAuth(BaseZitiAuth):
    requires_response_body = True

    async def async_auth_flow(
        self, request: httpx.Request
    ) -> AsyncGenerator[httpx.Request, httpx.Response]:
        request.headers["zt-session"] = self.api.token or ""
        response = yield request

        if response.status_code == 401:  # pragma: no branch
            refresh_request = self.build_refresh_request()
            refresh_response = yield refresh_request
            await refresh_response.aread()
            self.update_token(refresh_response)
            request.headers["zt-session"] = self.api.token or ""
            yield request

    def build_refresh_request(self) -> httpx.Request:
        """Builds the token refresh request."""
        return httpx.Request(
            "POST",
            f"{self.api.base_url}/authenticate",
            params={"method": "password"},
            json={
                "username": self.api.settings.ziti.auth.username,
                "password": self.api.settings.ziti.auth.password,
            },
        )


class ZitiIdentityAuth(BaseZitiAuth):
    requires_response_body = True

    def __init__(self, client: BaseZitiAPI):
        super().__init__(client)
        self.identity_context = ZitiIdentityAuthContext(self.api.settings.ziti.auth.identity)

    async def async_auth_flow(
        self, request: httpx.Request
    ) -> AsyncGenerator[httpx.Request, httpx.Response]:
        request.headers["zt-session"] = self.api.token or ""
        response = yield request

        if response.status_code == 401:  # pragma: no cover
            # Use the new client certificate authentication method
            refresh_response = await self.get_auth_token()
            # await refresh_response.aread()
            self.update_token(refresh_response)
            request.headers["zt-session"] = self.api.token or ""
            yield request

    async def get_auth_token(self):
        async with httpx.AsyncClient(
            base_url=self.api.base_url,
            verify=self.identity_context.ssl_context,
        ) as client:
            response = await client.post(
                "/authenticate",
                params={"method": "cert"},
            )
            return response


class ZitiManagementAPI(BaseZitiAPI):
    @property
    def base_url(self):
        return f"{self.settings.ziti.api.management}/edge/management/v1"

    def services(self) -> AsyncGenerator[dict[str, Any], None]:
        return self.collection_iterator("/services")

    def identities(self) -> AsyncGenerator[dict[str, Any], None]:
        return self.collection_iterator("/identities")

    async def search_config(self, id_or_name) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/configs", id_or_name)

    async def create_config(
        self, name: str, config_type_id: str, tags: TagsType | None = None
    ) -> str:
        return await self.create(
            "/configs",
            {
                "configTypeId": config_type_id,
                "name": name,
                "data": {
                    "auth_scheme": "none",
                    "basic_auth": None,
                    "interstitial": True,
                    "oauth": None,
                },
            },
            tags,
        )

    async def delete_config(self, config_id: str) -> None:
        return await self.delete("/configs", config_id)

    async def create_config_type(self, name: str, tags: TagsType | None = None) -> str:
        return await self.create(
            "/config-types",
            {
                "name": name,
                "schema": {},
            },
            tags,
        )

    async def create_service(
        self,
        name: str,
        config_id: str,
        tags: TagsType | None = None,
    ) -> str:
        return await self.create(
            "/services",
            {
                "name": name,
                "configs": [config_id],
                "encryptionRequired": True,
            },
            tags,
        )

    async def create_service_router_policy(
        self,
        name: str,
        service_id: str,
        tags: TagsType | None = None,
    ) -> str:
        return await self.create(
            "/service-edge-router-policies",
            {
                "name": name,
                "edgeRouterRoles": ["#all"],
                "serviceRoles": [
                    f"@{service_id}",
                ],
                "semantic": "AllOf",
            },
            tags,
        )

    async def create_router_policy(
        self,
        name: str,
        identity_id: str,
        tags: TagsType | None = None,
    ) -> str:
        return await self.create(
            "/edge-router-policies",
            {
                "name": name,
                "edgeRouterRoles": ["#all"],
                "identityRoles": [f"@{identity_id}"],
                "semantic": "AllOf",
            },
            tags,
        )

    async def search_service_router_policy(self, id_or_name: str) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/service-edge-router-policies", id_or_name)

    async def search_router_policy(self, id_or_name: str) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/edge-router-policies", id_or_name)

    async def delete_service_router_policy(self, policy_id: str) -> None:
        return await self.delete("/service-edge-router-policies", policy_id)

    async def delete_router_policy(self, policy_id: str) -> None:
        return await self.delete("/edge-router-policies", policy_id)

    async def search_service(self, id_or_name: str) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/services", id_or_name)

    async def get_service(self, service_id: str) -> dict[str, Any]:
        return await self.get("/services", service_id)

    async def delete_service(self, service_id: str) -> None:
        return await self.delete("/services", service_id)

    async def create_user_identity(self, name: str, tags: TagsType | None = None) -> str:
        return await self._create_identity(name, "User", tags=tags)

    async def create_device_identity(self, name: str, tags: TagsType | None = None) -> str:
        return await self._create_identity(name, "Device", tags=tags)

    async def search_identity(self, id_or_name: str) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/identities", id_or_name)

    async def search_config_type(self, id_or_name: str) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/config-types", id_or_name)

    async def get_identity(self, identity_id: str) -> dict[str, Any]:
        return await self.get("/identities", identity_id)

    async def delete_identity(self, identity_id: str) -> None:
        return await self.delete("/identities", identity_id)

    async def fetch_ca_certificates(self) -> str:
        response = await self.httpx_client.get("/.well-known/est/cacerts")
        response.raise_for_status()
        return response.text

    async def create_dial_service_policy(
        self, name: str, service_id: str, identity_id: str, tags: TagsType | None = None
    ) -> str:
        return await self._create_service_policy("Dial", name, service_id, identity_id, tags)

    async def create_bind_service_policy(
        self, name: str, service_id: str, identity_id: str, tags: TagsType | None = None
    ) -> str:
        return await self._create_service_policy("Bind", name, service_id, identity_id, tags)

    async def search_service_policy(self, id_or_name: str) -> dict[str, Any] | None:
        return await self.search_by_id_or_name("/service-policies", id_or_name)

    async def delete_service_policy(self, policy_id: str) -> None:
        return await self.delete("/service-policies", policy_id)

    async def _create_service_policy(
        self,
        type: Literal["Dial", "Bind"],
        name: str,
        service_id: str,
        identity_id: str,
        tags: TagsType | None = None,
    ) -> str:
        return await self.create(
            "/service-policies",
            {
                "name": name,
                "type": type,
                "serviceRoles": [f"@{service_id}"],
                "identityRoles": [f"@{identity_id}"],
                "semantic": "AllOf",
            },
            tags,
        )

    async def _create_identity(
        self,
        name: str,
        type: Literal["User", "Device", "Default"],
        tags: TagsType | None = None,
    ) -> str:
        return await self.create(
            "/identities",
            {
                "name": name,
                "type": type,
                "isAdmin": False,
                "enrollment": {"ott": True},
            },
            tags,
        )


class ZitiClientAPI(BaseZitiAPI):
    @property
    def base_url(self):
        return f"{self.settings.ziti.api.client}/edge/client/v1"

    async def enroll_identity(self, jti: str, csr_pem: str) -> dict[str, Any]:
        response = await self.httpx_client.post(
            "/enroll",
            params={
                "method": "ott",
                "token": jti,
            },
            headers={"Content-Type": "application/x-pem-file"},
            content=csr_pem,
        )
        response.raise_for_status()
        return response.json()



import pytest
from pytest_mock import MockerFixture

from mrok.ziti.bootstrap import bootstrap_identity
from tests.conftest import SettingsFactory


@pytest.mark.asyncio
async def test_bootstrap_identity(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocked_mgmt_api = mocker.AsyncMock()

    mocked_mgmt_api.search_identity.return_value = None
    mocker.patch(
        "mrok.ziti.bootstrap.enroll_proxy_identity",
        new_callable=mocker.AsyncMock,
        return_value=("identity-id", {"id": "key, cert and ca", "ztAPI": "https://ziti.api"}),
    )
    mocked_mgmt_api.search_router_policy.return_value = None
    mocked_mgmt_api.create_router_policy.return_value = "router-id"
    mocked_mgmt_api.search_config_type.return_value = None
    mocked_mgmt_api.create_config_type.return_value = "config-type-id"

    identity_id, identity_json = await bootstrap_identity(
        mocked_mgmt_api,
        mocker.AsyncMock(),
        settings.proxy.identity,
        settings.proxy.mode,
        False,
        None,
    )

    assert identity_id == "identity-id"
    assert identity_json == {"id": "key, cert and ca", "ztAPI": "https://ziti.api"}


@pytest.mark.asyncio
async def test_bootstrap_identity_with_forced(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocked_mgmt_api = mocker.AsyncMock()

    mocked_mgmt_api.search_identity.return_value = {"id": "old-proxy-identity-id"}
    mocked_mgmt_api.search_router_policy.side_effect = [
        {"id": "old-router-policy-id", "name": "policy"},
        None,
    ]
    mocked_mgmt_api.delete_router_policy.return_value = None
    mocked_mgmt_api.delete_identity.return_value = None

    mocker.patch(
        "mrok.ziti.bootstrap.enroll_proxy_identity",
        new_callable=mocker.AsyncMock,
        return_value=("new-identity-id", {"id": "key, cert and ca", "ztAPI": "https://ziti.api"}),
    )

    mocked_mgmt_api.create_router_policy.return_value = "new-policy-id"
    mocked_mgmt_api.search_config_type.return_value = None
    mocked_mgmt_api.create_config_type.return_value = "config-type-id"

    identity_id, identity_json = await bootstrap_identity(
        mocked_mgmt_api,
        mocker.AsyncMock(),
        settings.proxy.identity,
        settings.proxy.mode,
        True,
        {},
    )

    assert identity_id == "new-identity-id"
    assert identity_json == {"id": "key, cert and ca", "ztAPI": "https://ziti.api"}


@pytest.mark.asyncio
async def test_bootstrap_identity_with_forced_and_no_policy(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocked_mgmt_api = mocker.AsyncMock()

    mocked_mgmt_api.search_identity.return_value = {"id": "old-proxy-identity-id"}
    mocked_mgmt_api.search_router_policy.return_value = None
    mocked_mgmt_api.create_router_policy.return_value = "policy-id"
    mocked_mgmt_api.delete_identity.return_value = None

    mocker.patch(
        "mrok.ziti.bootstrap.enroll_proxy_identity",
        new_callable=mocker.AsyncMock,
        return_value=("new-identity-id", {"id": "key, cert and ca", "ztAPI": "https://ziti.api"}),
    )

    mocked_mgmt_api.search_config_type.return_value = None
    mocked_mgmt_api.create_config_type.return_value = "config-type-id"

    identity_id, identity_json = await bootstrap_identity(
        mocked_mgmt_api,
        mocker.AsyncMock(),
        settings.proxy.identity,
        settings.proxy.mode,
        True,
        {},
    )

    assert identity_id == "new-identity-id"
    assert identity_json == {"id": "key, cert and ca", "ztAPI": "https://ziti.api"}


@pytest.mark.asyncio
async def test_bootstrap_identity_already_bootstrapped(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocked_mgmt_api = mocker.AsyncMock()

    mocked_mgmt_api.search_identity.return_value = {"id": "identity-id"}
    mocked_mgmt_api.search_router_policy.return_value = {
        "id": "router-policy-id",
        "name": "policy",
    }
    mocked_mgmt_api.search_config_type.return_value = {
        "id": "config-type-id",
        "name": "config-type",
    }

    identity_id, identity_json = await bootstrap_identity(
        mocked_mgmt_api,
        mocker.AsyncMock(),
        settings.proxy.identity,
        settings.proxy.mode,
        False,
        {},
    )

    assert identity_id == "identity-id"
    assert identity_json is None
